/**
 * Copyright (C) 2011 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <sys/stat.h>

#define PAM_SM_AUTH

#include <security/pam_modules.h>
#include <security/_pam_macros.h>
#include <security/pam_ext.h>

#define DEFAULT_FILE "/cdrom/tickets"
#define DEFAULT_USER "ticketuser"

char *file = NULL;
char *user = NULL; /* not used ATM */
int use_first_pass = 0;
int try_first_pass = 0;

void
_pam_parse (const pam_handle_t *pamh, int argc, const char **argv)
{
    for (; argc-- > 0; ++argv)
    {
        if ((strlen (*argv) >5) && (strncmp (*argv, "file=", 5) == 0))
            user = strdup ((*argv)+5);
        if ((strlen (*argv) >5) && (strncmp (*argv, "user=", 5) == 0))
            file = strdup ((*argv)+5);
        else
            pam_syslog(pamh, LOG_ERR, "unknown option: %s", *argv);
    }

    if (! file)
        file = strdup (DEFAULT_FILE);
    if (! user)
        user = strdup (DEFAULT_USER);
}

PAM_EXTERN int pam_sm_authenticate (pam_handle_t *pamh, int flags,
                                    int argc, const char **argv)
{
    int retval = PAM_AUTH_ERR;
    char *resp = NULL;
    FILE *rf, *wf;
    char *tmpfile = NULL;
    char line[64];
    int found = 0;

    _pam_parse (pamh, argc, argv);

    retval = pam_set_item(pamh, PAM_USER, (const void *) user);
    if (retval != PAM_SUCCESS)
    {
        pam_syslog (pamh, LOG_ERR, "unable to set user");
        return PAM_USER_UNKNOWN;
    }

    retval = pam_prompt (pamh, PAM_PROMPT_ECHO_OFF, &resp,
                         "Ticket: ");
    if (retval != PAM_SUCCESS)
    {
        goto end;
    }
    pam_set_item (pamh, PAM_AUTHTOK, resp);

    if (!resp)
        goto end;

    if ((rf = fopen (file, "r")) == NULL)
        goto end;

    tmpfile = (char *) malloc (strlen (file) + 5);
    if (! tmpfile)
    {
        fclose (rf);
        goto end;
    }
    sprintf (tmpfile, "%s.tmp", file);

    if ((wf = fopen (tmpfile, "w")) == NULL)
    {
        fclose (rf);
        goto end;
    }

    retval = PAM_AUTH_ERR;
    while (fgets (line, 63, rf) != NULL)
    {
        if (line[strlen (line) - 1] == '\n')
            line[strlen (line) - 1] = '\0';

        if (!found && strncmp (line, resp, 63) == 0)
        {
            retval = PAM_SUCCESS;
            continue;
        }

        fprintf (wf, "%s\n", line);
    }

    fclose (rf);
    fclose (wf);

    chmod (tmpfile, 0600);
    if (rename (tmpfile, file) == -1)
        printf ("%s\n", strerror (errno));

end:
    if (resp)
    {
        _pam_overwrite (resp);
        _pam_drop (resp);
    }

    if (user) free (user);
    if (file) free (file);
    if (tmpfile) free (tmpfile);

    return retval;
}

PAM_EXTERN int
pam_sm_setcred (pam_handle_t *pamh, int flags,
                int argc, const char **argv)
{
    return PAM_IGNORE;
}

#ifdef PAM_STATIC
struct pam_module _pam_ticket_modstruct = {
    "pam_ticket",
    pam_sm_authenticate,
    pam_sm_setcred,
    NULL,
    NULL,
    NULL,
    NULL,
};
#endif
